@@ -119,6 +119,20 @@ function isOfType(obj, type) {
     return !isNull(obj) && obj.constructor===type;
 }
 
+/**
+ * Destroys the given object with C++ ownership.
+ */
+function destr(obj) {
+    if (!RSettings.isQt(6)) {
+        obj.destroy();
+    }
+    else {
+        // destroy does not work in Qt 6 / QJSEngine:
+        // ("Invalid attempt to destroy() an indestructible object")
+        obj.destr();
+    }
+}
+
 /**
  * Checks if the given object is undefined or null.
  *
@@ -443,6 +457,15 @@ function isEllipseEntity(obj) {
     return isOfType(obj, REllipseEntity) || isOfType(obj, REllipseEntityPointer);
 }
 
+/**
+ * Checks if the given object is an ellipse entity and represents an ellipse arc (not a full ellipse).
+ *
+ * \return true if the given object is an ellipse arc entity (REllipseEntity).
+ */
+function isEllipseArcEntity(obj) {
+    return (isOfType(obj, REllipseEntity) || isOfType(obj, REllipseEntityPointer)) && !obj.isFullEllipse();
+}
+
 /**
  * Checks if the given object is an ellipse entity and represents a full ellipse.
  *
@@ -525,6 +548,18 @@ function isClosedPolylineEntity(obj, tolerance) {
     }
 }
 
+/**
+ * \return true if the given object is a geometrically open polyline entity.
+ */
+function isOpenPolylineEntity(obj, tolerance) {
+    if (isNull(tolerance)) {
+        return isPolylineEntity(obj) && !obj.isGeometricallyClosed();
+    }
+    else {
+        return isPolylineEntity(obj) && !obj.isGeometricallyClosed(tolerance);
+    }
+}
+
 /**
  * Checks if the given object is a solid entity.
  *
@@ -868,6 +903,8 @@ function entityTypeToClass(type) {
         return "RLinetype";
     case RS.ObjectView:
         return "RView";
+    case RS.ObjectDimStyle:
+        return "RDimStyle";
 
     case RS.EntityAll:
         return "REntity";
@@ -969,6 +1006,8 @@ function entityTypeToString(type, plural) {
         return plural ? qsTr("Linetypes") : qsTr("Linetype");
     case RS.ObjectView:
         return plural ? qsTr("Views") : qsTr("View");
+    case RS.ObjectDimStyle:
+        return plural ? qsTr("Dimension Styles") : qsTr("Dimension Style");
 
     case RS.EntityAll:
         return plural ? qsTr("Entities") : qsTr("Entity");
@@ -1505,6 +1544,23 @@ Array.prototype.containsIgnoreCase = function(obj) {
     return this.contains(obj, function(a,b) { return a.toUpperCase()===b.toUpperCase(); });
 };
 
+/**
+ * \return True if this array contains all items of the given array.
+ */
+Array.prototype.containsAll = function(obj) {
+    if (obj.length>this.length) {
+        return false;
+    }
+
+    for (var i=0; i<obj.length; i++) {
+        if (this.indexOf(obj[i])<0) {
+            return false;
+        }
+    }
+
+    return true;
+};
+
 /**
  * \return Array with all elements of this array that are not also
  * in the given array (subtract).
@@ -1630,6 +1686,42 @@ Array.alphaNumericalSorter = function(a, b) {
     return RS.compareAlphanumerical(a, b);
 };
 
+Array.prototype.filter = function(func, thisArg) {
+    if ( ! ((typeof func === 'Function' || typeof func === 'function') && this) )
+        throw new TypeError();
+
+    var len = this.length >>> 0,
+        res = new Array(len), // preallocate array
+        t = this, c = 0, i = -1;
+
+    var kValue;
+    if (thisArg === undefined){
+      while (++i !== len){
+        // checks to see if the key was set
+        if (i in this){
+          kValue = t[i]; // in case t is changed in callback
+          if (func(t[i], i, t)){
+            res[c++] = kValue;
+          }
+        }
+      }
+    }
+    else{
+      while (++i !== len){
+        // checks to see if the key was set
+        if (i in this){
+          kValue = t[i];
+          if (func.call(thisArg, t[i], i, t)){
+            res[c++] = kValue;
+          }
+        }
+      }
+    }
+
+    res.length = c; // shrink down array to proper size
+    return res;
+};
+
 /**
  * Sorts this array with string items (case insensitive)
  */
@@ -1639,6 +1731,22 @@ Array.alphaNumericalSorter = function(a, b) {
 //    });
 //};
 
+Array.union = function(x, y) {
+    var i;
+    var obj = {};
+    for (i=x.length-1; i>=0; --i) {
+        obj[x[i]] = x[i];
+    }
+    for (i=y.length-1; i>=0; --i) {
+        obj[y[i]] = y[i];
+    }
+    var res = []
+    for (var k in obj) {
+        res.push(obj[k]);
+    }
+    return res;
+};
+
 /* Finds the intersection of two sorted arrays.
  *
  * \param a first array, must already be sorted
@@ -1888,7 +1996,7 @@ String.prototype.wordWrap = function(m, c, b){
 String.prototype.regexIndexOf = function(regex, startpos) {
     var indexOf = this.substring(startpos || 0).search(regex);
     return (indexOf >= 0) ? (indexOf + (startpos || 0)) : indexOf;
-}
+};
 
 String.prototype.regexLastIndexOf = function(regex, startpos) {
     regex = (regex.global) ? regex : new RegExp(regex.source, "g" + (regex.ignoreCase ? "i" : "") + (regex.multiLine ? "m" : ""));
@@ -1905,7 +2013,7 @@ String.prototype.regexLastIndexOf = function(regex, startpos) {
         regex.lastIndex = ++nextStop;
     }
     return lastIndexOf;
-}
+};
 
 /**
  * Modulo which works also for negative numbers (workaround for
@@ -2367,6 +2475,7 @@ function initUserShortcuts() {
         if (isNull(scStringList)) {
             if (match) qDebug("explicitly no shortcuts");
             action.setShortcuts([]);
+            action.removeShortcuts();
             continue;
         }
 
@@ -2380,6 +2489,7 @@ function initUserShortcuts() {
             var sc = new QKeySequence(scStringList[k]);
             scList.push(sc);
         }
+        action.removeShortcuts();
         action.setShortcuts(scList);
     }
 
@@ -2410,6 +2520,7 @@ function addActionsToWidgets() {
 
     // delete previous shortcut delegates:
     // part of workaround for QTBUG-38256, QTBUG-57990:
+    /*
     if ((RSettings.getQtVersion()<0x050500 && RSettings.getQtVersion()>=0x050000) ||
         (RSettings.getQtVersion()===0x050800 && RS.getSystemId()==="linux")) {
 
@@ -2424,9 +2535,10 @@ function addActionsToWidgets() {
         }
         appWin.setProperty("DelegatedShortcutsObjs", []);
     }
+    */
 
     var actions = RGuiAction.getActions();
-    var widgetTypes = ["Menu", "ToolBar", "MatrixPanel", "Panel"];
+    var widgetTypes = ["Menu", "ToolBar", "MatrixPanel", "Panel", "UserToolBar1", "UserToolBar2"];
     for (var c=0; c<actions.length; ++c) {
         var a = actions[c];
         if (a.icon.isNull() && !a.isIconDisabled()) {
@@ -2444,6 +2556,7 @@ function addActionsToWidgets() {
 
             var visibility = true;
             if (wn[0]==="!") {
+                // action not visible by default in this widget:
                 visibility = false;
                 wn = wn.substring(1);
             }
@@ -2472,6 +2585,7 @@ function addActionsToWidgets() {
 
                 // workaround for QTBUG-38256 (action not triggered for letter based shortcuts in SUB menus):
                 // workaround for QTBUG-57990 (action not triggered for letter based shortcuts in ALL menus):
+                /*
                 if ((RSettings.getQtVersion()<0x050500 && RSettings.getQtVersion()>=0x050000) ||
                     (RSettings.getQtVersion()===0x050800 && RS.getSystemId()==="linux")) {
 
@@ -2502,6 +2616,7 @@ function addActionsToWidgets() {
                         a.setDefaultShortcuts([]);
                     }
                 }
+                */
 
                 if (visibility) {
                     RGuiAction.addToWidget(a, w);
@@ -2522,7 +2637,9 @@ function addActionsToWidgets() {
             }
         }
 
-        if (!addedToWidget) {
+        if ((RSettings.getQtVersion()<0x050500 && RSettings.getQtVersion()>=0x050000) ||
+            (RSettings.getQtVersion()===0x050800 && RS.getSystemId()==="linux") ||
+            !addedToWidget) {
             // always add action to main window to make sure keycodes
             // are active even if action is invisible:
             RGuiAction.addToWidget(a, appWin);
@@ -2646,15 +2763,15 @@ function autoIconPath(path) {
 }
 
 function applyTheme() {
+    var systemId = RS.getSystemId();
     var theme = RSettings.getValue("Theme/ThemeName", undefined);
-    if (!isNull(theme)) {
+    if (systemId!=="osx" && !isNull(theme)) {
         var path = "themes/" + theme + "/";
 
         qApp.styleSheet = "";
 
         // load stylesheet.css, stylesheet_[win|macos|linux].css:
         var found = false;
-        var systemId = RS.getSystemId();
         if (systemId==="osx") systemId = "macos";
 
         var postfixes = ["", "_" + systemId];
@@ -2771,6 +2888,49 @@ function qsTranslate2(context, sourceText, disambiguation, n) {
     return qsTranslate(context, sourceText, disambiguation, n);
 }
 
+function makeQDirFilters() {
+    var argumentsNew = [].slice.call(arguments, 0);
+    argumentsNew.unshift(QDir.Filters);
+    return makeFlags.apply(null, argumentsNew);
+}
+
+function makeQDirSortFlags() {
+    var argumentsNew = [].slice.call(arguments, 0);
+    argumentsNew.unshift(QDir.SortFlags);
+    return makeFlags.apply(this, argumentsNew);
+}
+
+function makeQtMatchFlags() {
+    var argumentsNew = [].slice.call(arguments, 0);
+    argumentsNew.unshift(Qt.MatchFlags);
+    return makeFlags.apply(this, argumentsNew);
+}
+
+function makeFlags() {
+    if (RSettings.isQt(6)) {
+        var ret = 0;
+        for (var i=1; i<arguments.length; i++) {
+            ret |= arguments[i];
+        }
+        return ret;
+    }
+    else {
+        var argumentsNew = [].slice.call(arguments, 0);
+        var flgs = argumentsNew.shift();
+        return flgs.apply(this, argumentsNew);
+    }
+}
+
+function getQLocaleName(locale) {
+    if (isFunction(locale.name)) {
+        return locale.name();
+    }
+    else {
+        // QJSEngine does not allow "name" as function.
+        return locale.name2();
+    }
+}
+
 // fix QPlainTextEdit API for Qt 5:
 if (!isFunction(QPlainTextEdit.prototype.toPlainText)) {
     QPlainTextEdit.prototype.toPlainText = function() {
